/*------------------------------------------------------------------------------------------*/
/*This is the lowlevel IEC-bus_driver, 														*/
/*																							*/
/*Note that this driver is dirty because it polls the drive for it's flags. The 			*/
/*routines in this driver should therefore only be called from a low priority task			*/
/*                                                                                          */
/*Attention: routines tuned for 20MHz                                                       */
/*------------------------------------------------------------------------------------------*/
/*  History:                                                                        
    -------- 
	2008-09-21	removed IEC_monitor... it was never used (succesfully)     
	2008-03-10	adjusted the routines IEC_send_string and IEC_send_number_as_ASCII to handle the abort function (when IEC_send returns false, it should stop)
	2008-02-28	started with the IEC-abort functionality (when the user presses runs/stop during load)
	2007-04-28	KeepItSimpleStupid, so hanging on to this live style I removed the 'filter' from the IEC_receive routine (it was unneccessary nonsense anyway)
	2007-04-25	CheckForCommand had an delay of 200uS to ensure the "frame to release of ATN'-time" of min 20uS, this is removed and now in the higher level IEC staemachine (main.c)
	2007-04-23	IECundoturnaround() was checking for the clok, this is not reuiqed in this situation, that's why the timeout worked bettre then no timeout, simply because the CLOCK check is not required at THIS stage of the communication
	2007-04-16	removed a semicolon in the acknowledge timeout check... there should not be a semicolon after the while :-) (oops)
	2007-04-15	ADDED a ATN check on IEC_turnaround routine
	2007-04-09	removed all code that was made commentary with //
	2007-04-08	added the same 'filter' on FreeIEC and CheckForCommand... and then disabled the new code with "//" simply because the new improvements made it even worse...
	2007-04-07	added some extra checks to make sure we trigger/test the clock/data signals correctly (software signal filtering).
				this is required because sometimes the device triggers incorrectly on signals meant for other devices on the bus
				which is best noticable during intense bus trafic. But it is very likely to happen at any moment (you can't predict glitches/spikes).
	2007-04-01	the check IEC_CLOCK == 1 in UNDOTURNAROUND now has a time-out, simply to improve reliablillity... it helps a lot!
	2006-11-01	the channel value was always extracted from the commandbyte, even if there was no channel value. This caused problems in the new statemachine (before the new statemachine the channel was used before the 3f/5f command so this problme was no problem then)
	2006-10-31	added thefunction IEC_monitor this function can be used to monitor bus activity without interfering
				removed the setdeviceaddress function, since this is handled in MAIN.C
	2006-10-30	the data and open command are identical when the errorchannel (15) is addressed
	2006-10-22	changed the channel-variable to a variable outside the IEC_bus code, so it can be used in MAIN.c where it is declared
	2006-10-20	added a function to set the timings of the IEC-bus (required for 25% faster VIC20-mode)
	2006-07-03	changed 'FreeIEC' routine, this routine caused busylooping at an unwanted level, the busy-loop waiting for ATN going low 
				is now located in main, so we can do other things while not communicating over the IEC-bus (manual file selection for instance)
	2006-07-02	changed WaitForCommand into CheckForCommand, since the routine is no longer a while loop
	2006-03-28	change in release and pulldown ATN-line routines, caused by new PCB design
	2006-02-22	added frame_handshake signal for later use
	2006-02-02	improvements in code layout... i.o.w. making it more readable                                                                  
	2005-10-18	start of IEC bus driver
	2005-10-19	further details on EOI condition
	2005-10-22	delay routines for IEC timing
	2005-10-25	corrected the IEC-receive routine (without EOI detection)
	2005-10-26	timeouts and EOI handling
	2005-10-27	bus turnaround routines and testing of write routines. Optimisation of the send routines
	2006-01-04	added IEC_send_string, is much more convinient then printf
	2006-01-10	TALK and LISTEN definitions were mixed up, is corrected. FreeIEC-routine added and improved
	2006-01-13	adjusted delay in waitforcommand routine, now external devices work
	2006-01-16	rewritten the waitforcommand routine
*/

/*  TO DO:                                                                          
    ------      
	- EOI acknowledge delay, set/test this to correct timings for 20MHz

*/

/*------------------------------------------------------------------------------------------*/

/*--------------------------------------------------------*/
/*                        includes                        */
/*--------------------------------------------------------*/
#include <pic18.h>
#include <stdio.h>
#include <IEC_bus.h>
#include <delay.h>
#include <hardware.h>

/*--------------------------------------------------------*/
/*                        constants                       */   
/*--------------------------------------------------------*/
#define	FALSE			0							/*FALSE*/
#define	TRUE			1							/*TRUE*/

#define JIFFY_OFFSET_SEND 3		/* timing constant for centering the sampling window */
#define JIFFY_OFFSET_RECV 3

/*--------------------------------------------------------*/
/*                         globals                        */
/*--------------------------------------------------------*/
unsigned char			IEC_receive_delay;			/*set delay to ..uSec*/
unsigned char			IEC_send_delay;				/*set delay to ..uSec*/

/*external variables (variables also used in other .c files)*/
/*external variables (variables also used in other .c files)*/
extern bit 				EOI;						/*End Of Indicator: this flag (when TRUE) indicates that the last byte is being transmitted or received*/
extern bit 				TimeOut;					/*Timeout-flag*/

extern bit 				jiffy_active;				/*JiffyDos active flag: 1 = active */
extern bit				jiffy_load;					/*1 = JiffyDos LOAD detected, 0 = non-LOAD case */
extern bit				jiffy_ATN;					/* used to store the ATN line state */
extern bit				jiffy_CLK;					/* used to store the CLK line state */
extern bit				FC3_active;					/*FC3 fastloader flag: 1 = active */

extern unsigned char	channel;					/*this register holds the channel value*/

extern unsigned char	devicenumber;				/* device number */

/*--------------------------------------------------------*/
/*                     local functions                    */
/*--------------------------------------------------------*/


/*--------------------------------------------------------*/
/*                    external functions                  */
/*--------------------------------------------------------*/

//IEC_ATN : when IEC_ATN=0 it means that all devices must switch to listen (receive) mode immediatly


/*set all IEC_signal lines in the correct mode*/
void InitIEC(void)
{
	SetIEC_timings(TIMINGS_DEFAULT);	/*set timings according the default settings (C64 compatible mode)*/
	IEC_atn_release();
 	IEC_clock_release();
	IEC_data_release();
	EOI = FALSE;
	TimeOut = FALSE;				

	jiffy_active = 0;	// reset JiffyDOS flags too
	jiffy_load = 0;

}

void SetIEC_timings(unsigned char settings)
{
	switch(settings)
	{
		case TIMINGS_VIC20:				/*Timings for VIC-20 mode*/
		{								/*-----------------------*/
			IEC_receive_delay = 52;		/*set delay to ..uSec*/
			IEC_send_delay = 60;		/*set delay to ..uSec*/
			break;
		}

		case TIMINGS_DTV:				/*Timings for DTV (tuned IEC-bus speeds)*/
		{								/*--------------------------------------*/
			IEC_receive_delay = 60;		/*set delay to ..uSec*/
			IEC_send_delay = 80;		/*set delay to ..uSec*/
			break;
		}

		default:						/*Timings according (C64) IEC-specs*/
		{								/*---------------------------------*/
			IEC_receive_delay = 70;		/*set delay to ..uSec*/
			IEC_send_delay = 80;		/*set delay to ..uSec*/
			break;
		}		
	}
}

/*free the IEC bus for use by other devices, the bus is free'd until ATN is released*/
void FreeIEC(void)
{
	IEC_atn_release();
 	IEC_clock_release();
	IEC_data_release();
	while(IEC_ATN == 0);
}


/*this routine will wait untill it receives one byte over the IEC-bus*/
unsigned char IEC_receive(unsigned char frame_handshake)
{
	unsigned char data;
	unsigned char bitcount;
	unsigned long t;

	EOI = FALSE;							/*clear the EOI flag*/
	TimeOut = FALSE;						/*clear the timeout flag*/
	data = 0;								/*clear destination register*/
	bitcount = 8;							/*the number of bits within a byte are still eight*/
	t=0;
	

	//OutputToRS232(); // debug only

	/*-----Step1:Ready to send--------------------------------------------------------------------------------*/
	while (IEC_CLOCK == 0)					/*wait (endlessly) until we see some action on the bus, we may respond whenever we please*/
	{
		if (IEC_ATN == 1)					/*check for JiffyDOS bus 'garbage' prior to real command, ATN rises before clock to signal something! */
		{
			TimeOut = TRUE;
			return(0);						/*exit receiving loop*/
		}	
	}


	/*Step2:Ready for data--------------------------------------------------------------------------------*/
	IEC_data_release();						/*response to talker READY FOR DATA, this signal has no time-limit since we may be busy formatting a disk or printing out a chunk of text to paper*/

	/*-----intermission:EOI--------------------------------------------------------------------------------*/
	t = 5000;						/*timeout value for detecting an EOI condition*/
	while (IEC_CLOCK == 1)			/*wait for the talker to indicate that the transmission will start (this signal must be received within 200uS, otherwise we may consider it as an EOI)*/
	{
		t--;						/*when timed out, we have detected an EOI condition*/
		if (t == 0)					/*timeout detected ?*/
		{
			EOI = TRUE;					/*set the EOI flag*/
			IEC_data_pulldown();		/*acknowledge the EOI condition*/
			DelayBigUs(IEC_receive_delay);	/*delay for .. uSec*/
			IEC_data_release();			/*acknowledge the EOI condition*/
		}
	}

	/*-----Step3:receiving the bits--------------------------------------------------------------------------------*/
	while (bitcount != 0)			/*data is received with LSB first*/
	{
		data>>=1;					/*shift the byte one bit to the right*/
		bitcount--;					/*decrement bitcounter by one*/
	
		if (bitcount == 0)			/*if we're at bit number 7, JiffyDos causes a 500us delay between transmission of clock */
		{
			t=100;				 // load counter with JiffyDOS detection time (about 300us read from LA)
			
			while (IEC_CLOCK == 0)	 
			{
				t--;					
				if ( (t == 0) && (IEC_ATN == 0) ) // check if timeout occurred and we are reading a command (ATN = 0)!
				{
					if ( (data&0x0F) == devicenumber )	/*check if data is meant for us */
					{
						IEC_data_pulldown();		/*notify master that we support JD too by toggling data line!*/
						DelayBigUs(101);		
						IEC_data_release();
						jiffy_active = 1;	/*set JiffyDos detected flag */
						t = 1000; // reload counter, not really necessary
					}
				} // if t=0
			} // while IEC_CLOCK == 0	

		} // if bitcount == 0

		else						/*proceed normally with timeout detection in other bits*/ 
		{
			t = 5000;					/*timeout value*/
			while (IEC_CLOCK == 0)		/*wait for the talker to release the clock line to signal "bit ready"*/
			{
				t--;					/*timeout detection is required to prevent "hanging" in case of data transmission erros*/
				if (t == 0)				/*timeout detected ?*/
				{
					TimeOut = TRUE;		/*set the timeout flag*/
					return(0);			/*exit*/
				}
			} // while	
		} // else

		if (IEC_DATA == 1)			/*test if bit is set or cleared (attention:documentation states that signals are inverted ?!?! error ?!?!*/
			data = data + 128;		/*make bit 7 high*/

		t = 5000;					/*timeout value*/
		while (IEC_CLOCK == 1)		/*wait for the talker to pulldown the clock line*/
		{
			t--;					/*timeout detection is required to prevent "hanging" in case of data transmission erros*/
			if (t == 0)				/*timeout detected ?*/
			{
				TimeOut = TRUE;		/*set the timeout flag*/
				return(0);			/*exit*/
			}
		}		
	} // receiving the bits

	/*-----Step4:frame handshake--------------------------------------------------------------------------------*/
	if (frame_handshake == FALSE)	/*we must acknowledge within 1 mS that we have received the data, if not the talker knows there is something wrong*/
		DelayBigUs(1250);			/*delay for more then 1000 uSec*/		

	DelayBigUs(25);					// small delay inserted to make it look more like LA bus traces ;-)
	IEC_data_pulldown();			
	
	//OutputToLCD();	// debug only!

	return(data);					/*exit with the received data*/
}


/*this routine will send one byte over the IEC-bus*/
unsigned char IEC_send(unsigned char data)
{
	unsigned char bitcount;
	unsigned long t;
	bitcount = 8;					/*the number of bits within a byte are still eight*/
	TimeOut = FALSE;				/*clear the timeout flag*/

	//OutputToRS232();
	//printf("\r\n%02x",data);
	//OutputToLCD();

	/* send a JiffyDOS byte with the non-LOAD protocol */
	/* jiffy_send(data, eoi, loadflags) */
	/* jiffy_send returns the negative state of IEC_ATN */
	if (jiffy_active == 1)
	{
		/* Checking of ATN transition is needed to get by JLOAD address bytes, otherwise always a ?LOAD ERROR */
		/* JLOAD interrupts the transfer after the two address bytes are transfered */

		if (IEC_ATN == 0)
			return(FALSE);

		DelayBigUs(260); // short delay between bytes is needed, 360us is the same as in JD1541

		/* Checking of ATN transition is needed to get by JLOAD address bytes */
		if (IEC_ATN == 0)
			return(FALSE);

		if( EOI == TRUE ) // check if we need to send the JD non-LOAD EOI mark
		{
			jiffy_send(data, 1, 0); // send the non-LOAD EOI mark after the final data byte
			return (TRUE);
		}	

		if( jiffy_send(data, 0, 0) == 1 ) // send data byte and check if IEC_ATN was low when exiting jiffy_send, something wrong?
		{			
			if (IEC_ATN == 0)
				return(FALSE); /*exit*/
			return (FALSE);
		}	

		DelayUs(100);	/* Wait for ATN toggle after re-entrant calls, return FALSE if it occurs */
		if(IEC_ATN==0)
			return(FALSE);
		else
			return (TRUE); // exit, byte transferred successfully!
	}

	/* Send a normal IEC byte if JiffyDOS is not found */

	/*-----Check for IEC-abort--------------------------------------------------------------------------------*/
	/*IEC abort: the computer sends an ATN during data transfer from device to computer, */
	/*the device sees that this is a special situation and aborts, freeing up the bus and */
	/*the computer will get it's prompt back (test this by loading a program and pressing run/stop during loading)*/
	/*(this information was supplied by: Dirk (a.k.a. SKERN, from www.dienstagstreff.de)*/

	if (IEC_ATN == 0)
	{
		return(FALSE);				/*exit*/		
	}
	
	/*-----Step1:Ready to send--------------------------------------------------------------------------------*/

	IEC_clock_release();			/*indicate to the listener that we are ready to send*/

	/*-----Step2:Ready for data--------------------------------------------------------------------------------*/
	while (IEC_DATA == 0);			/*wait for the listener (computer) to release the data-line (ready to receive data), this may take forever... (according the IEC-bus definition)*/

	/*intermission:EOI--------------------------------------------------------------------------------*/
	if (EOI == TRUE)
	{
		DelayUs(200);				/*delay for more then 200 uSec*/
		t = 5000;					/*timeout value*/
		while (IEC_DATA == 1)		/*check for acknowledge of the EOI. The acknowledge is simply pulling the data-line down for a brief period*/
		{
			t--;					/*timeout detection is required to prevent "hanging" in case of data transmission erros*/
			if (t == 0)				/*timeout detected ?*/
			{
				TimeOut = TRUE;		/*set the timeout flag*/
				return(FALSE);		/*exit*/
			}
		}
		t = 5000;					/*timeout value*/
		while (IEC_DATA == 0)		/**/
		{
			t--;					/*timeout detection is required to prevent "hanging" in case of data transmission erros*/
			if (t == 0)				/*timeout detected ?*/
			{
				TimeOut = TRUE;		/*set the timeout flag*/
				return(FALSE);	   	/*exit*/
			}
		}
		/*EOI is acknowledged by the listener (computer)*/
	}
	IEC_clock_pulldown();			/*indicate to the listener (computer) that the transmission will start*/
	DelayUs(IEC_send_delay);		/*delay for .. uSec*/

	/*-----Step3:receiving the bits--------------------------------------------------------------------------------*/
	while (bitcount != 0)			/*data is transmitted with LSB first*/
	{
		IEC_clock_pulldown();		/*release the clockline in order to set the next bit on the data-line*/
		DelayUs(IEC_send_delay);	/*delay for .. uSec*/
		bitcount--;
		if (data & 0x01)
		{
			IEC_data_release();		
		}
		else
		{		
			IEC_data_pulldown();			
		}
		IEC_clock_release();		/*indicate to the listener (computer) "bit ready"*/
		data>>=1;					/*shift the byte one bit to the right*/
		DelayUs(IEC_send_delay);	/*delay for at least 60uSec when a C64 is listening*/
	}
	IEC_clock_pulldown();			/*the whole byte has been transmitted*/
	DelayUs(IEC_send_delay);		/*delay for at least 60uSec when a C64 is listening*/
	IEC_data_release();				/*set clock and data in the correct order and wait for an acknowledge*/
	DelayUs(25);					/*wait a few uSec to prevent a false trigger*/
	
	/*----Step4:frame handshake--------------------------------------------------------------------------------*/
	t = 5000;						/*timeout value*/
	while (IEC_DATA == 1)			/*wait for the listener (computer) to acknowledge this byte it must respond within 1 mSec*/
	{
			t--;					/*timeout detection is required to prevent "hanging" in case of data transmission erros*/
			if (t == 0)				/*timeout detected ?*/
			{
				TimeOut = TRUE;		/*set the timeout flag*/
				return(FALSE);		/*exit*/
			}
	}		
	DelayUs(100);					/*time between bytes is according IEC-specs at least 100 uSec*/
	return(TRUE);					/*the byte has been succesfully transmitted*/
}


/*this routine will handle strings, for sending strings over the IEC-bus*/
unsigned char IEC_send_string(const unsigned char *inputstring)
{
	unsigned char lp;

	lp = 0;
	while(inputstring[lp] != 0)
	{	
		if(IEC_send(inputstring[lp]) == FALSE)
			return(FALSE);
		lp++;
	}
	return(TRUE);
}

unsigned char IEC_send_number_as_ASCII(unsigned char number)
{
	unsigned char i;

	i = 0x30 + (number/10);			/*10's of the value*/
	if(IEC_send(i) == FALSE)
		return(FALSE);
	i = 0x30 + (number%10);			/*1's of the value*/
	if(IEC_send(i) == FALSE)
		return(FALSE);

	return(TRUE);
}

/*this routine will reverse the direction of the bus*/
void IEC_turnaround(void)
{
	while (IEC_ATN == 0);			/*wait (endlessly) until the computer releases ATN*/

	//if (jiffy_active == 1)
	//{
	//	IEC_data_release();
	//	DelayUs(50);
	//	IEC_clock_pulldown();
	//	DelayUs(70);
	//}
	//else
	//{
		while (IEC_CLOCK == 0);			/*wait (endlessly) until the computer releases the clock line*/
		IEC_data_release();
		DelayUs(25);					/*small delay... to make the timing look like the timing of a real drive*/
		IEC_clock_pulldown();
		DelayUs(160);					/*Tda = 80 uSec minimum acknowledge time for acknowledging the device is now a talker*/
	//}
	//OutputToRS232();
	//printf("\r\n\Turnaround");
	//OutputToLCD();
}


/*this routine will set the direction on the bus back to normal (the way it was when the computer was switched on)*/
void IEC_undoturnaround(void)
{
//	unsigned int timeout_cnt;
//	DelayUs(50);					/*Talk-Attention release (max. 100uS)*/

	if  (jiffy_active==1)
	{
		IEC_data_pulldown();
		IEC_clock_release();
		DelayUs(25);					/*small delay for .. uSec in order to prevent a false trigger*/
	}
	else
	{
		IEC_data_pulldown();
		IEC_clock_release();
		DelayUs(25);					/*small delay for .. uSec in order to prevent a false trigger*/
	}
//	timeout_cnt = 0;
//	while (IEC_CLOCK == 1)			/*wait until the computer pulls down the clock line or we timeout*/
//	{
//		if (timeout_cnt == 1000)
//			break;					/*timeout exceeded, leave routine*/
//		else
//			timeout_cnt++;
//	}
}

/*this routine will let the bus lines go, as Jim Butterfield described, both talker and listener 'letgo'*/
void IEC_letgo(void)
{
	DelayUs(50);					/*after a suitable pause*/
	IEC_data_release();				/*dataline and clockline are released*/
	IEC_clock_release();
}


unsigned char CheckForCommand(unsigned char *outputcommand, unsigned char frame_handshake)
{
	if(IEC_ATN == 1)				/* wait untill there is relevant info (a command is indicated by the attention signal) on the bus */
	{
		IEC_clock_release();		/* added by J. Saily 19.4.2009, not sure if needed? */
		IEC_data_release();			/*free the bus... (if not allready released)*/
		return(FALSE);				/*no command received, yet...*/
	}

	IEC_data_pulldown();			/*join the bus...*/

	DelayUs(55);	/* Delay to wait for CLK to go down after ATN, this may take upto 20us on a JD-C64! */

	*outputcommand = IEC_receive(frame_handshake);	/*save the received command to a register specified by the caller of this routine*/
	if ((TimeOut == FALSE) && (IEC_ATN == 0))
	{

		if (*outputcommand == TALK)
			return(TALK);

		if (*outputcommand == LISTEN)
			return(LISTEN);

		if (*outputcommand == UNLISTEN)				/*check for command:UNLISTEN, device:..., Attention-line must still be active*/
			return(UNLISTEN);
	
		if (*outputcommand == UNTALK)				/*check for command:UNTALK, device:..., Attention-line must still be active*/
			return(UNTALK);

		if(*outputcommand == JDATA) 				/*check for JiffyDOS LOAD condition! 0x61 */
		{
			 return(JDATA);
		}	

		if ((*outputcommand & 0xF0) == DATA)		/*check for command:DATA, channel:..., Attention-line must still be active*/
		{
			channel = *outputcommand & 0x0F;				/*(if send) the channel is stored in the low-nibble of the command*/		
			return(DATA);
		}

		if ((*outputcommand & 0xF0) == CLOSE)		/*check for command:CLOSE, channel:..., Attention-line must still be active*/
		{
			channel = *outputcommand & 0x0F;				/*(if send) the channel is stored in the low-nibble of the command*/		
			return(CLOSE);
		}

		if ((*outputcommand & 0xF0) == OPEN)		/*check for command:OPEN, channel:..., Attention-line must still be active*/
		{
			channel = *outputcommand & 0x0F;				/*(if send) the channel is stored in the low-nibble of the command*/		
			return(OPEN);
		}

		return(*outputcommand);						/*when none of the above... it's for another device, return the entire command for further analysis*/
	}
	return(FALSE);									/*code should never reach this point!!!*/
}


/*--------------------------------------------------------*/
/*                      local functions                   */
/*--------------------------------------------------------*/

void IEC_atn_release(void)
{
	IEC_ATN = 1;			/*set pin high, usefull when we want to set it to an output later*/
	TRISB = TRISB | 0x01;	/*set pin (RB0) to input*/
}

void IEC_atn_pulldown(void)
{
	TRISB = TRISB & 0xFF;	/*set pin (RB0) to output*/
	IEC_ATN = 0;			/*set pin low*/
}

void IEC_clock_release(void)
{
	IEC_CLOCK = 1;			/*set pin high, usefull when we want to set it to an output later*/
	TRISA = TRISA | 0x04;	/*set pin (RA2) to input*/
}

void IEC_clock_pulldown(void)
{
	TRISA = TRISA & 0xFB;	/*set pin (RA2) to output*/
	IEC_CLOCK = 0;			/*set pin low*/
}

void IEC_data_release(void)
{
	IEC_DATA = 1;			/*set pin high, usefull when we want to set it to an output later*/
	TRISA = TRISA | 0x08;	/*set pin (RA3) to input*/
}

void IEC_data_pulldown(void)
{
	TRISA = TRISA & 0xF7;	/*set pin (RA3) to output*/
	IEC_DATA = 0;			/*set pin low*/
}



unsigned char jiffy_receive(unsigned char frame_handshake) 
{
  unsigned char data,tmp;

	data = 0;
  
    /* Set clock+data high */
    IEC_clock_release();	//set_clock(1);
    IEC_data_release();		//set_data(1);

    /* Wait until clock is high and emulate ATN-Ack */
    while (!IEC_CLOCK) 
	{
      if (!IEC_ATN)
        IEC_data_pulldown();	//set_data(0);
    }

	DelayUs(13 + JIFFY_OFFSET_RECV);

	/* Bits 5+4 */
	if(IEC_DATA == 1)
		data = data | 0x20; // bit 5: 0010 0000
	if(IEC_CLOCK == 1)
		data = data | 0x10; // bit 4: 0001 0000

	DelayUs(11); // instruction clock period is 0.2us, estimate 10 cycles for setting data bits?

    /* Bits 7+6 */
 	if(IEC_DATA == 1)
		data = data | 0x80; // bit 7: 1000 0000
	if(IEC_CLOCK == 1)
		data = data | 0x40; // bit 6: 0100 0000

	DelayUs(11);

    /* Bits 1+3 */
	if(IEC_DATA == 1)
		data = data | 0x02; // bit 1: 0000 0010
	if(IEC_CLOCK == 1)
		data = data | 0x08; // bit 3: 0000 1000

	DelayUs(11);

    /* Bits 0+2 */
	if(IEC_DATA == 1)
		data = data | 0x01; // bit 0: 0000 0001
	if(IEC_CLOCK == 1)
		data = data | 0x04; // bit 2: 0000 0100

	DelayUs(11);

    /* Read EOI mark */
    //start_timeout(6);
    //*busstate = IEC_PIN;
    //while (!has_timed_out()) ;

	DelayUs(6);

	// add checks here with global flags for:
	// - if IEC_ATN was active (low) at the end of transfer? If so, then receive a normal IEC-command next.
	// - if IEC_CLOCK was high, meaning EOI received
	// modify WaitForString function in main.c accordingly!

	if (IEC_ATN == 1)
		jiffy_ATN = 1;
	else
		jiffy_ATN = 0;

	if (IEC_CLOCK == 1)
		jiffy_CLK = 1;
	else
		jiffy_CLK = 0;

    /* Data low */
    IEC_data_pulldown();

	return (data ^ 0xff);	// invert the received data byte
}



unsigned char jiffy_send(unsigned char value, unsigned char eoi, unsigned char loadflags) 
{
  unsigned char waitcond, eoimark, tmp;

	//IEC_DATA = RA3
	//IEC_CLOCK = RA2
	//IEC_ATN = RB0

	di();	// disable all interrupts during Jiffy-transfers

	/* modified from sd2iec code base */
	/* if eoi = 1 then send non-LOAD eoi marker */
	/* if loadflags = 1 then send using LOAD protocol (eoi is handled differently, and not sent in this function! */

	/* Using separate load routines for non-LOAD and JLOAD cases, for timing reasons! */
	/* FIX ME: combine non-LOAD and JLOAD routines to save memory space! */
	
	/* non-LOAD case is handled here*/
	/* non-LOAD is used mostly for directory loading! */
	if (loadflags == 0)
	{
		tmp = PORTA; // load the PORTA values into tmp
	   
		TRISA |= 0b00001100; // set IEC_CLOCK and IEC_DATA to input (and release to high), this starts a JiffyDOS symbol
		PORTA |= 0b00001100; // raise both IEC_CLOCK and IEC_DATA outputs to prevent glitches in transmit when TRISA is changed next!
		while( IEC_ATN==1 && IEC_CLOCK==1 && IEC_DATA==0 ); // wait condition for symbol synchronization
		TRISA = TRISA & 0b11110011;	/*set pins (RA2) and (RA3) to output*/

    	/* first bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		//DelayUs(1);	// wait for the sampling window 6-13us from zero (was 1)
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;

    	/* second bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(15); // 10us melkein toimii (10)
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;
    	
    	/* third bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(7); // 7us melkein toimii
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;

    	/* fourth bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(7); // 7
		PORTA = tmp; // output the clock & data sequence;

    	tmp = 0;
    	/* EOI marker for non-LOAD routine */
    	if (eoi == 1) 	/* if eoi = 1 then send EOI marker after byte! */
		{
  			DelayUs(11);	
			PORTA &= 0b11110111;	// pull IEC_data low		
			PORTA |= 0b00000100;	// raise IEC_clock high
			TRISA |= 0b00000100;	// set IEC_clock to input
      		DelayUs(1);	// let bus stabilize
      		tmp = !IEC_ATN;   // store the (negative) state of IEC_ATN line
			ei(); // enable interrupts
			return (tmp);
    	}
  	
		DelayUs(11); // 11us
		PORTA |= 0b00001000;	// raise IEC_data high
		PORTA &= 0b11111011;	// pull IEC_clock low
		TRISA |= 0b00001000;	// set IEC_data to input
		DelayUs(1); // let bus stabilize
		while( (IEC_ATN==1) && (IEC_DATA==1) ); // wait until C64 acknowledges transmission by pulling the data low (or toggles ATN to interrupt)
  		tmp = !IEC_ATN;   // store the (negative) state of IEC_ATN line
		ei();	// enable interrupts
		return (tmp);
	} // end non-LOAD routine

/* loadflags != 0, so a JLOAD transfer is handled */
/* loadflags = 1 means a normal JLOAD byte */
/* loadflags = 2 means a wait for next buffer condition signaled by IEC_CLOCK low before the first bitpair */
	if (loadflags == 1)	
	{
		tmp = PORTA; // load the PORTA values into tmp

	    //IEC_data_release(); 	//set_data(1) and input ;
  	 	//IEC_clock_release();	//set_clock(1) and input;    this is the start of a JiffyDOS symbol

		//TRISA = TRISA & 0b11111111;	/*set pins (RA2) and (RA3) to input and start JD symbol */
		TRISA |= 0b00001100; // set IEC_CLOCK and IEC_DATA to input

		PORTA |= 0b00001100; // raise both IEC_CLOCK and IEC_DATA outputs to prevent glitches in transmit when TRISA is changed next!

		while( IEC_ATN==1 && IEC_CLOCK==1 && IEC_DATA==1 ); // wait condition for symbol synchronization

		TRISA = TRISA & 0b11110011;	/*set pins (RA2) and (RA3) to output*/

    	/* first bitpair */
  		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(1);	// wait for the sampling window 6-13us from zero (1)
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;

    	/* second bitpair */
    	tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(14); // 10us after first (10)
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;

    	/* third bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(9); // 11us after second (7)
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;
 
    	/* fourth bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(7); // 10us after third (7)
		PORTA = tmp; // output the clock & data sequence;

    	tmp = 0;
  		tmp = !IEC_ATN;   // store the (negative) state of IEC_ATN line

		ei();	// enable interrupts

		return (tmp);
	}

/* loadflags = 2 means send a wait for next buffer condition signaled by IEC_CLOCK low after the final bitpair of a block buffer */
/* This path is also used in signaling the end-of-file together with the actual EOI marker in main code */
	if (loadflags == 2)	
	{
		tmp = PORTA; // load the PORTA values into tmp

	    //IEC_data_release(); 	//set_data(1) and input ;
  	 	//IEC_clock_release();	//set_clock(1) and input;    this is the start of a JiffyDOS symbol
  
		//TRISA = TRISA & 0b11111111;	/*set pins (RA2) and (RA3) to input and start JD symbol */
		TRISA |= 0b00001100; // set IEC_CLOCK and IEC_DATA to input

		PORTA |= 0b00001100; // raise both IEC_CLOCK and IEC_DATA outputs to prevent glitches in transmit when TRISA is changed next!
		
		while( IEC_ATN==1 && IEC_CLOCK==1 && IEC_DATA==1 ); // wait condition for symbol synchronization
		
		TRISA = TRISA & 0b11110011;	/*set pins (RA2) and (RA3) to output*/

    	/* first bitpair */
   		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(1);	// wait for the sampling window 6-13us from zero
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;

    	/* second bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(14); // 10us melkein toimii
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;

    	/* third bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(9); // 7us melkein toimii
		PORTA = tmp; // output the clock & data sequence;
    	value >>= 2;

    	/* fourth bitpair */
		tmp = tmp & 0b11110011; // make IEC_CLOCK and IEC_DATA zero for now
		if (value & 0x01)	/* bit 0: 0000 0001 */	
			tmp |= 0b00000100;	// raise IEC_CLOCK bit; IEC_clock_release();
		if (value & 0x02)	/* bit 1: 0000 0010 */
			tmp |= 0b00001000;	// raise IEC_DATA bit; IEC_data_release();
		DelayUs(7);
		PORTA = tmp; // output the clock & data sequence;

    	tmp = 0;

		DelayUs(11);

		/* Send end-of-block marker with clock low, data high*/
     	//IEC_clock_pulldown(); // signal the wait for next buffer mark by pulling IEC_CLOCK low
		//IEC_data_release();

		PORTA |= 0b00001000; // set IEC_DATA (high / inactive)
		PORTA &= 0b11111011; // clear IEC_CLOCK (low / active)
   		TRISA |= 0b00001000;	/*set IEC_DATA to input (IEC_CLOCK remains output) */
   		//DelayUs(1);

		/* Wait until Data and/or ATN are low */
   		while ( (IEC_ATN == 1) && (IEC_DATA == 1) );

		if(IEC_ATN ==0)
			return(1);	// return with inverse of ATN to signal a transmission error

		if(eoi==0)
		{
			/* Exit with the inverse ATN in tmp */
			/* C64 waits for the loading of next block from disk and a new start condition to mark the new block beginning, this is handled in DataHandler() */      		
			DelayUs(10); // wait a while to make sure C64 finishes checking the wait-for-next-block condition
			tmp = !IEC_ATN;   // store the (negative) state of IEC_ATN line
			ei();	// enable interrupts
			return (tmp);
		}
		else
		{
			/* Send JLOAD end-of-file marker */
			DelayUs(100);
			PORTA |= 0b00000100; // set IEC_CLOCK (high)
			DelayUs(100);
			PORTA &= 0b11111011; // clear IEC_CLOCK (low)
			DelayUs(100);
			PORTA |= 0b00000100; // set IEC_CLOCK (high);

			tmp = !IEC_ATN;   // store the (negative) state of IEC_ATN line
			ei();	// enable interrupts
			return (tmp);	
		}	
  	
		// code should never reach this point
		tmp = !IEC_ATN;   // store the (negative) state of IEC_ATN line
		ei();	// enable interrupts
		return (tmp);
	}

}


/*
The person who writes their sourcecode without comments is either very smart or very stupid
*/
